//////////////////////////////////////////////////////////////////////////////////////////
// Project description
// 
// Name: Touchlib barrel distortion correction tool
//
// Laurence Muller (aka Falcon4ever)
// Contact: info@falcon4ever.com
// Date: 19/12/2007
// Site: www.multigesture.net
// Licensetype: Same as the original sample of OpenCV
//
// Purpose: 
// - Correcting the barrel distortion of the lens
// - Supports usb and firewire camera's (CMU driver)
//
// This application is based on the example which can be found in:
// C:\Program Files\OpenCV\samples\c\calibration.cpp
//
// Other references:
// http://groups.google.mv/group/OpenCV/browse_thread/thread/371a93cf636f976c - Adding a border to fix the crop
// Touchlib CMU capture filter by Johannes Hirche (xvarg) - CMU driver support
//
// Instructions can be found in the readme.txt
//
// Todo:
// - Changing the border/crop dynamicly
// - Cleaning the source
// - Support for dsvl?
//
//////////////////////////////////////////////////////////////////////////////////////////

// Adjustable parameters
int _border_size = 0;	// 20 = good for a F 2.5mm lens, please set this to zero if not needed.
// Please set USB camera settings -> line 107 of this file
// Please set firewire camera settings (default values) -> cmu_init()


// Horrible sourcecode starts here.
#include "main.h"

int main( int argc, char** argv )
{	
	printf ("Barrel distortion correction tool v0.1\n\n");

	const char* live_capture_help = 
			"When the live video from camera is used as input, the following hot-keys may be used:\n"
            "  <ESC>, 'q' - quit the program\n"
            "  'g' - start capturing images\n"
            "  'u' - switch undistortion on/off\n";

	// Display arguments
	//-w 6 -h 8 -s 2 -n 10
	if( argc < 2 )
    {
        printf( "This is a camera calibration sample.\n"
            "Usage: calibration\n"
            "     -w <board_width>         # the number of inner corners per one of board dimension\n"
            "     -h <board_height>        # the number of inner corners per another board dimension\n"
            "     [-n <number_of_frames>]  # the number of frames to use for calibration\n"
            "                              # (if not specified, it will be set to the number\n"
            "                              #  of board views actually available)\n"
            "     [-d <delay>]             # a minimum delay in ms between subsequent attempts to capture a next view\n"
            "                              # (used only for video capturing)\n"
            "     [-s <square_size>]       # square size in some user-defined units (1 by default)\n"           
			"     [-b <border_size>]       # Add a black border to increase the resized image (0 by default)\n"         
            "\n" );        
        return 0;
    }
	
	// Read arguments
	for( i = 1; i < argc; i++ )
    {
        const char* s = argv[i];
        if( strcmp( s, "-w" ) == 0 )
        {
            if( sscanf( argv[++i], "%u", &board_size.width ) != 1 || board_size.width <= 0 )
                return fprintf( stderr, "Invalid board width\n" ), -1;
        }
        else if( strcmp( s, "-h" ) == 0 )
        {
            if( sscanf( argv[++i], "%u", &board_size.height ) != 1 || board_size.height <= 0 )
                return fprintf( stderr, "Invalid board height\n" ), -1;
        }
        else if( strcmp( s, "-s" ) == 0 )
        {
            if( sscanf( argv[++i], "%f", &square_size ) != 1 || square_size <= 0 )
                return fprintf( stderr, "Invalid board square width\n" ), -1;
        }
        else if( strcmp( s, "-n" ) == 0 )
        {
            if( sscanf( argv[++i], "%u", &image_count ) != 1 || image_count <= 3 )
                return printf("Invalid number of images\n" ), -1;
        }        
        else if( strcmp( s, "-d" ) == 0 )
        {
            if( sscanf( argv[++i], "%u", &delay ) != 1 || delay <= 0 )
                return printf("Invalid delay\n" ), -1;
        }
		else if( strcmp( s, "-b" ) == 0 )
        {
            if( sscanf( argv[++i], "%u", &_border_size ) != 1 || _border_size <= 0 )
                return printf("Invalid border size\n" ), -1;
        }
        else
            return fprintf( stderr, "Unknown option %s", s ), -1;
    }

	// Check camera type	
	printf ("Please select your camera type:\n");
	printf ("1. USB camera (uses OpenCV framegrabber)\n");
	printf ("2. Firewire camera (uses CMU framegrabber)\n");
	printf ("3. Exit\n\n");

	scanf ("%d", &interface_type);

	if(interface_type == 1)
	{		
		capture = cvCreateCameraCapture(0);
		cvSetCaptureProperty(capture, CV_CAP_PROP_FRAME_WIDTH, 640.);
		cvSetCaptureProperty(capture, CV_CAP_PROP_FRAME_HEIGHT, 480.);
		
		if( !capture )
		{
			printf("Could not initialize video capture\n");
			return 0;	
		}
	}
	else if(interface_type == 2)
	{
		if(cmu_init() == CAM_SUCCESS)
		{
			
			printf("Found a firewire camera!\n");
		}
		else
		{		
			printf("Smth went wrong, could not find a firewire camera.\n");
			return 0;		
		}
	}
	else
	{
		printf ("\nBye!\n");
		return 0;
	}


	if(capture || cmu_capture)
        printf( "\n%s", live_capture_help );

    

	elem_size = board_size.width * board_size.height * sizeof(image_points_buf[0]);
    storage = cvCreateMemStorage( MAX( elem_size*4, 1 << 16 ));
    image_points_buf = (CvPoint2D32f*)cvAlloc( elem_size );
    image_points_seq = cvCreateSeq( 0, sizeof(CvSeq), elem_size, storage );

    cvNamedWindow("Camera View", 1);

	for(;;)
    {
		IplImage *view = 0, *view_gray = 0;
        int count = 0, found, blink = 0;
        CvPoint text_origin;
        CvSize text_size = {0,0};
        int base_line = 0;
        char s[100];
        int key;

		if(interface_type == 1)
        {
            IplImage* view0 = cvQueryFrame( capture );

            if( view0 )
            {
                view = cvCreateImage( cvGetSize(view0), IPL_DEPTH_8U, view0->nChannels );
                if( view0->origin == IPL_ORIGIN_BL )
                    cvFlip( view0, view, 0 );
                else
                    cvCopy( view0, view );
            }
        }
		else if(interface_type == 2)
        {
			
			CvSize size;
			size.height = cmu_h;
			size.width = cmu_w;
			int channels = cmu_mono?1:3;
			IplImage* view0 = cvCreateImage(size, IPL_DEPTH_8U, channels);

			int dropped = 0;
			if(cmu_capture->AcquireImageEx(true, &dropped) == CAM_SUCCESS){
				if(cmu_mono){
					unsigned long len = view0->imageSize;
					memcpy(view0->imageData, cmu_capture->GetRawData(&len),len);
				}else{
					if(cmu_flipRGB){			
						cmu_capture->getRGB((unsigned char*)cmu_rawImage->imageData,cmu_rawImage->imageSize);
						cvCvtColor(cmu_rawImage,view0,CV_RGB2BGR); 
					}else{				
						cmu_capture->getRGB((unsigned char*)view0->imageData,view0->imageSize);					
					}
				}
			}

            if( view0 )
            {
                view = cvCreateImage( cvGetSize(view0), IPL_DEPTH_8U, view0->nChannels );
                if( view0->origin == IPL_ORIGIN_BL )
                    cvFlip( view0, view, 0 );
                else
                    cvCopy( view0, view );
            }
			
        }

if( !view )
        {
            if( image_points_seq->total > 0 )
            {
                image_count = image_points_seq->total;
                goto calibrate;
            }
            break;
        }

        if( flip_vertical )
            cvFlip( view, view, 0 );

        img_size = cvGetSize(view);
        found = cvFindChessboardCorners( view, board_size,
            image_points_buf, &count, CV_CALIB_CB_ADAPTIVE_THRESH );

#if 1
        // improve the found corners' coordinate accuracy
        view_gray = cvCreateImage( cvGetSize(view), 8, 1 );
        cvCvtColor( view, view_gray, CV_BGR2GRAY );
        cvFindCornerSubPix( view_gray, image_points_buf, count, cvSize(11,11),
            cvSize(-1,-1), cvTermCriteria( CV_TERMCRIT_EPS+CV_TERMCRIT_ITER, 30, 0.1 ));
        cvReleaseImage( &view_gray );
#endif

		
        if( mode == CAPTURING && found && (f || clock() - prev_timestamp > delay*1e-3*CLOCKS_PER_SEC) )		 
        {
            cvSeqPush( image_points_seq, image_points_buf );
            prev_timestamp = clock();
            blink = !f;
#if 1
            if(capture || cmu_capture)
            {
                sprintf( imagename, "view%03d.png", image_points_seq->total - 1 );
                cvSaveImage( imagename, view );
            }
#endif
        }
		 

        cvDrawChessboardCorners( view, board_size, image_points_buf, count, found );

        cvGetTextSize( "100/100", &font, &text_size, &base_line );
        text_origin.x = view->width - text_size.width - 10;
        text_origin.y = view->height - base_line - 10;

        if( mode == CAPTURING )
        {
            if( image_count > 0 )
                sprintf( s, "%d/%d", image_points_seq ? image_points_seq->total : 0, image_count );
            else
                sprintf( s, "%d/?", image_points_seq ? image_points_seq->total : 0 );
        }
        else if( mode == CALIBRATED )
            sprintf( s, "Calibrated" );
        else
            sprintf( s, "Press 'g' to start" );

        cvPutText( view, s, text_origin, &font, mode != CALIBRATED ?
                                   CV_RGB(255,0,0) : CV_RGB(0,255,0));

        if( blink )
            cvNot( view, view );

        if( mode == CALIBRATED && undistort_image )
        {
			
            IplImage* t = cvCloneImage( view );

			if(_border_size>0)
				view = undistorted_with_border( t, &camera, &dist_coeffs, 20 );	// seems good
			else
				cvUndistort2( t, view, &camera, &dist_coeffs );         
							
			cvReleaseImage( &t );
        }

        cvShowImage( "Camera View", view );
        key = cvWaitKey(capture ? 50 : 500);

        if( key == 27 )
            break;
        
        if( key == 'u' && mode == CALIBRATED )
            undistort_image = !undistort_image;

        if( capture && key == 'g' )
        {
            mode = CAPTURING;
            cvClearMemStorage( storage );
            image_points_seq = cvCreateSeq( 0, sizeof(CvSeq), elem_size, storage );
        }
		
		if( cmu_capture && key == 'g' )
        {
            mode = CAPTURING;
            cvClearMemStorage( storage );
            image_points_seq = cvCreateSeq( 0, sizeof(CvSeq), elem_size, storage );
        }

        if( mode == CAPTURING && (unsigned)image_points_seq->total >= (unsigned)image_count )
        {
calibrate:
            cvReleaseMat( &extr_params );
            cvReleaseMat( &reproj_errs );
            int code = run_calibration( image_points_seq, img_size, board_size,
                square_size, aspect_ratio, flags, &camera, &dist_coeffs, &extr_params,
                &reproj_errs, &avg_reproj_err );
            // save camera parameters in any case, to catch Inf's/NaN's

			save_camera_params(&camera, &dist_coeffs);
            if( code )
                mode = CALIBRATED;
            else
                mode = DETECTION;
        }

        if( !view )
            break;
        cvReleaseImage( &view );
	}

	if(capture)
	{
		cvReleaseCapture( &capture );		
	}

	if(cmu_capture)
	{
		cmu_capture->StopVideoStream();
		delete cmu_capture;
		cmu_capture = NULL;
	}

	return 0;
}

double compute_reprojection_error(	const CvMat* object_points,
									const CvMat* rot_vects, 
									const CvMat* trans_vects,
									const CvMat* camera_matrix, 
									const CvMat* dist_coeffs,
									const CvMat* image_points, 
									const CvMat* point_counts,
									CvMat* per_view_errors )
{
    CvMat* image_points2 = cvCreateMat( image_points->rows, image_points->cols, image_points->type );
    int i, image_count = rot_vects->rows, points_so_far = 0;
    double total_err = 0, err;
    
    for( i = 0; i < image_count; i++ )
    {
        CvMat object_points_i, image_points_i, image_points2_i;
        int point_count = point_counts->data.i[i];
        CvMat rot_vect, trans_vect;

        cvGetCols( object_points, &object_points_i,
            points_so_far, points_so_far + point_count );
        cvGetCols( image_points, &image_points_i,
            points_so_far, points_so_far + point_count );
        cvGetCols( image_points2, &image_points2_i,
            points_so_far, points_so_far + point_count );
        points_so_far += point_count;

        cvGetRow( rot_vects, &rot_vect, i );
        cvGetRow( trans_vects, &trans_vect, i );

        cvProjectPoints2( &object_points_i, &rot_vect, &trans_vect,
                          camera_matrix, dist_coeffs, &image_points2_i,
                          0, 0, 0, 0, 0 );
        err = cvNorm( &image_points_i, &image_points2_i, CV_L1 );
        if( per_view_errors )
            per_view_errors->data.db[i] = err/point_count;
        total_err += err;
    }
    
    cvReleaseMat( &image_points2 );
    return total_err/points_so_far;
}


int run_calibration(	CvSeq* image_points_seq, 
						CvSize img_size, 
						CvSize board_size,
						float square_size, 
						float aspect_ratio, 
						int flags,
						CvMat* camera_matrix, 
						CvMat* dist_coeffs, 
						CvMat** extr_params,
						CvMat** reproj_errs, 
						double* avg_reproj_err )
{
    int code;
    int image_count = image_points_seq->total;
    int point_count = board_size.width*board_size.height;
    CvMat* image_points = cvCreateMat( 1, image_count*point_count, CV_32FC2 );
    CvMat* object_points = cvCreateMat( 1, image_count*point_count, CV_32FC3 );
    CvMat* point_counts = cvCreateMat( 1, image_count, CV_32SC1 );
    CvMat rot_vects, trans_vects;
    int i, j, k;
    CvSeqReader reader;
    cvStartReadSeq( image_points_seq, &reader );

    // initialize arrays of points
    for( i = 0; i < image_count; i++ )
    {
        CvPoint2D32f* src_img_pt = (CvPoint2D32f*)reader.ptr;
        CvPoint2D32f* dst_img_pt = ((CvPoint2D32f*)image_points->data.fl) + i*point_count;
        CvPoint3D32f* obj_pt = ((CvPoint3D32f*)object_points->data.fl) + i*point_count;

        for( j = 0; j < board_size.height; j++ )
            for( k = 0; k < board_size.width; k++ )
            {
                *obj_pt++ = cvPoint3D32f(j*square_size, k*square_size, 0);
                *dst_img_pt++ = *src_img_pt++;
            }
        CV_NEXT_SEQ_ELEM( image_points_seq->elem_size, reader );
    }

    cvSet( point_counts, cvScalar(point_count) );

    *extr_params = cvCreateMat( image_count, 6, CV_32FC1 );
    cvGetCols( *extr_params, &rot_vects, 0, 3 );
    cvGetCols( *extr_params, &trans_vects, 3, 6 );

    cvZero( camera_matrix );
    cvZero( dist_coeffs );

    if( flags & CV_CALIB_FIX_ASPECT_RATIO )
    {
        camera_matrix->data.db[0] = aspect_ratio;
        camera_matrix->data.db[4] = 1.;
    }

    cvCalibrateCamera2( object_points, image_points, point_counts,
                        img_size, camera_matrix, dist_coeffs,
                        &rot_vects, &trans_vects, flags );

    code = cvCheckArr( camera_matrix, CV_CHECK_QUIET ) &&
        cvCheckArr( dist_coeffs, CV_CHECK_QUIET ) &&
        cvCheckArr( *extr_params, CV_CHECK_QUIET );

    *reproj_errs = cvCreateMat( 1, image_count, CV_64FC1 );
    *avg_reproj_err =
        compute_reprojection_error( object_points, &rot_vects, &trans_vects,
            camera_matrix, dist_coeffs, image_points, point_counts, *reproj_errs );

    cvReleaseMat( &object_points );
    cvReleaseMat( &image_points );
    cvReleaseMat( &point_counts );

    return code;
}


void save_camera_params(const CvMat* camera_matrix, CvMat* dist_coeffs)
{
	const char* configfile = "camera.yml";

    CvFileStorage* fs = cvOpenFileStorage( configfile, 0, CV_STORAGE_WRITE );
        
    cvWrite( fs, "camera_matrix", camera_matrix );
    cvWrite( fs, "distortion_coefficients", dist_coeffs );

    cvReleaseFileStorage( &fs );
}

int cmu_init()
{
	cmu_flipRGB = false;
	cmu_sMode = CMU640x480mono;
	cmu_sRate = CMURate30fps;

	cmu_mode = 5;
	cmu_rate = 4;
	cmu_format = 0;
	cmu_capture = NULL;
	cmu_lID.QuadPart = 0;
	cmu_mono = false;
	cmu_sharpness = 80;
	cmu_gain = 87;
	cmu_brightness = -1;
	cmu_exposure = -1;
	cmu_whitebalanceL = 0;
	cmu_whitebalanceH = -1;
	cmu_hue = -2;
	cmu_saturation = 90;
	cmu_gamma = 1;

	cmu_capture = new C1394Camera();
	cmu_capture->RefreshCameraList();
	int count = cmu_capture->GetNumberCameras();	
	if(!count){
		cmu_capture = NULL;
		return CAM_ERROR;
	}

	char *buffer = new char[255];
	int lastWorking = -1;
	bool searchCam = false;
	if(cmu_sVendor.size() || cmu_sName.size() || cmu_lID.QuadPart)
		searchCam = true;

	bool gotCam = false;
	for(int i=0;i<count;i++){				
		if(cmu_capture->SelectCamera(i) == CAM_SUCCESS){
			lastWorking = i;
			if(searchCam){
				if(cmu_lID.QuadPart){
					LARGE_INTEGER cID;
					cmu_capture->GetCameraUniqueID(&cID);					
					if(cID.QuadPart == cmu_lID.QuadPart){
						gotCam = true;
						break;
					}
				}
				if(cmu_sName.size()){
					cmu_capture->GetCameraName(buffer,255);
					buffer = _strlwr(buffer);
					if(cmu_sName == buffer){
						gotCam = true;
						break;
					}					
				}
				if(cmu_sVendor.size()){
					cmu_capture->GetCameraVendor(buffer,255);
					buffer = _strlwr(buffer);
					if(cmu_sVendor == buffer){
						gotCam = true;
						break;
					}					
				}
			}else{
				gotCam = true;
				break;
			}
		}
	}	
	delete buffer;
	if(!gotCam && lastWorking != -1){
		cmu_capture->SelectCamera(lastWorking);
	}else{
		if(!gotCam)
			goto bailout;
	}
	if(cmu_capture->InitCamera(true) != CAM_SUCCESS)
		goto bailout;
#ifdef _DEBUG
	printf("CMU Camera init succeeded\n");
#endif
	if(cmu_capture->SetVideoMode(cmu_mode) != CAM_SUCCESS)
		goto bailout;
#ifdef _DEBUG
	printf("CMU SetVideoMode succeeded, desired: %d, acquired: %d\n",cmu_mode,cmu_capture->GetVideoMode());
#endif
	if(cmu_capture->SetVideoFormat(cmu_format) != CAM_SUCCESS)
		goto bailout;
#ifdef _DEBUG
	printf("CMU SetVideoFormat succeeded, desired: %d, acquired: %d\n",cmu_format,cmu_capture->GetVideoFormat());
#endif
	if(cmu_capture->SetVideoFrameRate(cmu_rate) != CAM_SUCCESS)
		goto bailout;
#ifdef _DEBUG
	printf("CMU SetVideoFrameRate succeeded, desired: %d, acquired: %d\n",cmu_rate,cmu_capture->GetVideoFrameRate());
#endif
	cmu_capture->GetVideoFrameDimensions(&cmu_w,&cmu_h);
#ifdef _DEBUG
	printf("CMU VideoFrameDimensions: %d %d\n",cmu_w,cmu_h);
#endif

	CvSize size;
	size.height = cmu_h;
	size.width = cmu_w;
	int channels = cmu_mono?1:3;

	cmu_rawImage = cvCreateImage(size,IPL_DEPTH_8U, channels);
	cmu_capture->StartImageAcquisition();

	// grab some frames to clear out any junk
	for(int i=0;i<10;i++)
		cmu_capture->AcquireImageEx(true,0);

	// set all the weird stuff
	cmu_setFeature(FEATURE_BRIGHTNESS,cmu_brightness,0);
	cmu_setFeature(FEATURE_SHARPNESS,cmu_sharpness,0);
	cmu_setFeature(FEATURE_WHITE_BALANCE,cmu_whitebalanceL,cmu_whitebalanceH);
	cmu_setFeature(FEATURE_HUE,cmu_hue,0);
	cmu_setFeature(FEATURE_SATURATION,cmu_saturation,0);
	cmu_setFeature(FEATURE_GAMMA,cmu_gamma,0);
	cmu_setFeature(FEATURE_GAIN,cmu_gain,0);
	cmu_setFeature(FEATURE_AUTO_EXPOSURE,cmu_exposure,0);

	return CAM_SUCCESS;

	bailout:
	if(cmu_capture){
		delete cmu_capture;
		cmu_capture = NULL;
	}
	return CAM_ERROR;
}

void cmu_setFeature(CAMERA_FEATURE id, int valueL, int valueH)
{
	if(valueL == -2)
		return;

	if(cmu_capture->HasFeature(id)){
		C1394CameraControl *c = cmu_capture->GetCameraControl(id);
		if(c){
			if(valueL == -1){
				c->SetAutoMode(true);
			}else{
				c->SetAutoMode(false);
				c->SetValue(valueL,valueH);
			}
		}
	}
}

IplImage* undistorted_with_border( const IplImage *image, const CvMat *intrinsic,const CvMat *distortion, short int border )
{
    assert( image );
    assert( intrinsic && distortion );
    assert( border >= 0 );

    //add border to cx,cy
    CvMat *b_intrinsic = cvCloneMat( intrinsic );
    cvSetReal2D( b_intrinsic, 0,2, cvGetReal2D(b_intrinsic,0,2)+border );
    cvSetReal2D( b_intrinsic, 1,2, cvGetReal2D(b_intrinsic,1,2)+border );

    //add border to image
    IplImage *bordered = cvCreateImage( cvSize(image->width+2*border,image->height+2*border), image->depth, image->nChannels );
    cvCopyMakeBorder( image, bordered, cvPoint(border,border), IPL_BORDER_CONSTANT, cvScalarAll(0) );

    //undistort
    IplImage *bordered_corr = cvCreateImage( cvSize(image->width+2*border,image->height+2*border), image->depth, image->nChannels );
    cvUndistort2( bordered, bordered_corr, b_intrinsic, distortion );

    cvReleaseImage( &bordered );
    return bordered_corr;

}
